# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku CLI commands for handling of datasets.

Description
~~~~~~~~~~~

Create, edit and manage the datasets in your Renku project.

This is a core feature of Renku. You might want to go through the examples
listed below to get an idea of how you can create, import, and edit datasets.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.dataset:dataset
   :prog: renku dataset
   :nested: full

Examples
~~~~~~~~

Create an empty dataset inside a Renku project:

.. image:: ../../_static/asciicasts/dataset-create.delay.gif
   :width: 850
   :alt: Create a Dataset

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset create <dataset>
   :description: Create a new dataset called <dataset>.
   :target: rp,ui

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset ls
   :description: List all datasets in the project.
   :target: rp,ui

You can select which columns to display by using ``--columns`` to pass a
comma-separated list of column names:

.. code-block:: console

    $ renku dataset ls --columns id,name,date_created,creators
    ID        NAME           CREATED              CREATORS
    --------  -------------  -------------------  ---------
    0ad1cb9a  some-dataset   2020-03-19 16:39:46  sam
    9436e36c  my-dataset     2020-02-28 16:48:09  sam

Displayed results are sorted based on the value of the first column.

You can specify output formats by passing ``--format`` with a value of ``tabular``,
``json-ld`` or ``json``.

Showing dataset details:

.. code-block:: console

    $ renku dataset show some-dataset
    Name: some-dataset
    Created: 2020-12-09 13:52:06.640778+00:00
    Creator(s): John Doe<john.doe@example.com> [SDSC]
    Keywords: Dataset, Data
    Annotations:
    [
      {...}
    ]
    Title: Some Dataset
    Description:
    Just some dataset

You can also show details for a specific tag using the ``--tag`` option.

Deleting a dataset:

.. code-block:: console

    $ renku dataset rm some-dataset
    OK


.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset rm <dataset>
   :description: Remove a dataset.
   :target: rp

Creating a dataset with a storage backend:

By passing a storage URI with the ``--storage`` option, you can tell Renku that
the data for the dataset is stored in a remote storage. At the moment, Renku
supports only S3 backends. For example:

.. code-block:: console

    $ renku dataset create s3-data --storage s3://bucket-name/path

Renku prompts for your S3 credentials and can store them for future uses.

.. note:: Data directory for datasets that have a storage backend is ignored by
    Git. This is needed to avoid committing pulled data from a remote storage to Git.

Working with data
~~~~~~~~~~~~~~~~~

Adding data to the dataset:

.. image:: ../../_static/asciicasts/dataset-add.delay.gif
   :width: 850
   :alt: Add data to a Dataset

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset add <dataset> <url>
   :description: Add data from <url> to a dataset. <url> can be a local
                 file path, an http(s) address or a Git git+http or git+ssh repository.
   :target: rp,ui

This will copy the contents of ``data-url`` to the dataset and add it
to the dataset metadata.

.. note::

   If the URL refers to a local directory, data is added differently depending
   on if there is a trailing slash (``/``) or not. If the URL ends in a slash,
   files inside the directory are added to the target directory. If it does
   not end in a slash, then the directory itself will be added inside the
   target directory.

You can create a dataset when you add data to it for the first time by passing
``--create`` flag to add command:

.. code-block:: console

    $ renku dataset add --create new-dataset http://data-url

To add data from a git repository, you can specify it via https or git+ssh
URL schemes. For example,

.. code-block:: console

    $ renku dataset add my-dataset git+ssh://host.io/namespace/project.git

Sometimes you want to add just specific paths within the parent project.
In this case, use the ``--source`` or ``-s`` flag:

.. code-block:: console

    $ renku dataset add my-dataset --source path/within/repo/to/datafile \
        git+ssh://host.io/namespace/project.git

The command above will result in a structure like

.. code-block:: console

    data/
      my-dataset/
        datafile

You can use shell-like wildcards (e.g. *, **, ?) when specifying paths to be
added. Put wildcard patterns in quotes to prevent your shell from expanding
them.

.. code-block:: console

    $ renku dataset add my-dataset --source 'path/**/datafile' \
        git+ssh://host.io/namespace/project.git

You can use ``--destination`` or ``-d`` flag to set the location where the new
data is copied to. This location be will under the dataset's data directory and
will be created if it does not exists.

.. code-block:: console

    $ renku dataset add my-dataset \
        --source path/within/repo/to/datafile \
        --destination new-dir/new-subdir \
        git+ssh://host.io/namespace/project.git

will yield:

.. code-block:: console

    data/
      my-dataset/
        new-dir/
          new-subdir/
            datafile

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset add <dataset> --source <path>
             [--destination <rel-path>] <git-url>
   :description: Add only data in <path> from Git. With --destination:
                 location the data is copied to.
   :target: rp

To add a specific version of files, use ``--ref`` option for selecting a
branch, commit, or tag. The value passed to this option must be a valid
reference in the remote Git repository.

Updating a dataset:

After adding files from a remote Git repository or importing a dataset from a
provider like Dataverse or Zenodo, you can check for updates in those files by
using ``renku dataset update --all`` command. For Git repositories, this command
checks all remote files and copies over new content if there is any. It does
not delete files from the local dataset if they are deleted from the remote Git
repository; to force the delete use ``--delete`` argument. You can update to a
specific branch, commit, or tag by passing ``--ref`` option.
For datasets from providers like Dataverse or Zenodo, the whole dataset is
updated to ensure consistency between the remote and local versions. Due to
this limitation, the ``--include`` and ``--exclude`` flags are not compatible
with those datasets. Moreover, deleted remote files are automatically deleted
without requiring the ``--delete`` argument. Modifying those datasets locally
will prevent them from being updated.

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset update <dataset>
   :description: Update files in a dataset based on their source.
   :target: rp

The update command also checks for file changes in the project and updates
datasets' metadata accordingly. You can automatically add new files from
the dataset's data directory by using the ``--check-data-directory`` flag.

You can limit the scope of updated files by specifying dataset names, using
``--include`` and ``--exclude`` to filter based on file names, or using
``--creators`` to filter based on creators. For example, the following command
updates only CSV files from ``my-dataset``:

.. code-block:: console

    $ renku dataset update -I '*.csv' my-dataset

Note that putting glob patterns in quotes is needed to tell Unix shell not
to expand them.

You can use ``--dry-run`` flag to get a preview of what files/datasets will be
updated by an update operation.

Tagging a dataset:

A dataset can be tagged with an arbitrary tag to refer to the dataset at that
point in time. A tag can be added like this:

.. code-block:: console

    $ renku dataset tag my-dataset 1.0 -d "Version 1.0 tag"

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset tag <dataset> <tag> [-d <desc>]
   :description: Add a tag to the current version of the dataset, with
                 description <desc>.
   :target: rp

A list of all tags can be seen by running:

.. code-block:: console

    $ renku dataset ls-tags my-dataset
    CREATED              NAME    DESCRIPTION      DATASET     COMMIT
    -------------------  ------  ---------------  ----------  ----------------
    2020-09-19 17:29:13  1.0     Version 1.0 tag  my-dataset  6c19a8d31545b...

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset ls-tags <dataset>
   :description: List all tags for a dataset.
   :target: rp


A tag can be removed with:

.. code-block:: console

    $ renku dataset rm-tags my-dataset 1.0

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset rm-tags <dataset> <tags...>
   :description: Remove tags from a dataset.
   :target: rp


Importing data from other Renku projects:

To import all data files and their metadata from another Renku dataset use:

.. code-block:: console

    $ renku dataset import \
        https://renkulab.io/projects/<username>/<project>/datasets/<dataset-id>

or

.. code-block:: console

    $ renku dataset import \
        https://renkulab.io/projects/<username>/<project>/datasets/<dataset-name>

or

.. code-block:: console

    $ renku dataset import \
        https://renkulab.io/datasets/<dataset-id>

You can get the link to a dataset form the UI or you can construct it by
knowing the dataset's ID.

By default, Renku imports the latest version of a dataset from the other
project. If you want to import another version, pass the dataset version's tag
to the import command:

.. code-block:: console

    $ renku dataset import \
        https://renkulab.io/datasets/<dataset-id> --tag <version>


Importing data from an external provider:

.. image:: ../../_static/asciicasts/dataset-import.delay.gif
   :width: 850
   :alt: Import a Dataset

.. code-block:: console

    $ renku dataset import 10.5281/zenodo.3352150

This will import the dataset with the DOI (Digital Object Identifier)
``10.5281/zenodo.3352150`` and make it locally available.
Dataverse and Zenodo are supported, with DOIs (e.g. ``10.5281/zenodo.3352150``
or ``doi:10.5281/zenodo.3352150``) and full URLs (e.g.
``http://zenodo.org/record/3352150``). A tag with the remote version of the
dataset is automatically created.

You can change the directory a dataset is imported to by using the
``--datadir`` option.

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset import <uri>
   :description: Import a dataset. <uri> can be a Renku, Zenodo or Dataverse
                 URL or DOI.
   :target: rp

Exporting data to an external provider:

.. code-block:: console

    $ renku dataset export my-dataset zenodo

This will export the dataset ``my-dataset`` to ``zenodo.org`` as a draft,
allowing for publication later on. If the dataset has any tags set, you
can choose if the repository `HEAD` version or one of the tags should be
exported. The remote version will be set to the local tag that is being
exported.

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset export <dataset> <provider>
   :description: Export the dataset <dataset> to <provider>. Providers:
                 Zenodo, Dataverse.
   :target: rp

To export to a Dataverse provider you must pass Dataverse server's URL and
the name of the parent dataverse where the dataset will be exported to.
Server's URL is stored in your Renku setting and you don't need to pass it
every time.

To export a dataset to OLOS you must pass the OLOS server's base URL and
supply your access token when prompted for it. You must also choose which
organizational unit to export the dataset to from the list shown during
the export. The export does not map contributors from Renku to OLOS and
also doesn't map License information. Additionally, all file categories
default to Primary/Derived. This has to adjusted manually in the OLOS
interface after the export is done.


Exporting data to a local directory:

Renku provides a ``local`` provider that can be used to get a copy of a
dataset. For example, the following command creates a copy of the dataset
``my-dataset`` version ``v1`` in ``/tmp/my-dataset-v1``:

.. code-block:: console

    $ renku dataset export my-dataset local --tag v1 --path /tmp/my-dataset-v1

This also creates a copy of dataset's metadata at the given version and puts it
in ``<destination>/METADATA.yml``. If a destination path is not given to this
command, it creates a directory in project's data directory using dataset's
name and version: ``<data-dir>/<dataset-name>-<version>``. Export fails if the
destination directory is not empty.

.. note:: See our `dataset versioning tutorial
   <https://renkulab.io/projects/learn-renku/dataset-crates/dataset-versioning>`_
   for example recipes using tags for data management.

Listing all files in the project associated with a dataset.

.. code-block:: console

    $ renku dataset ls-files
    DATASET NAME         ADDED                PATH                           LFS
    -------------------  -------------------  -----------------------------  ----
    my-dataset           2020-02-28 16:48:09  data/my-dataset/add-me         *
    my-dataset           2020-02-28 16:49:02  data/my-dataset/weather/file1  *
    my-dataset           2020-02-28 16:49:02  data/my-dataset/weather/file2
    my-dataset           2020-02-28 16:49:02  data/my-dataset/weather/file3  *

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset ls-files
   :description: List all dataset files in project.
   :target: rp

You can select which columns to display by using ``--columns`` to pass a
comma-separated list of column names:

.. code-block:: console

    $ renku dataset ls-files --columns name,creators, path
    DATASET NAME         CREATORS   PATH
    -------------------  ---------  -----------------------------
    my-dataset           sam        data/my-dataset/add-me
    my-dataset           sam        data/my-dataset/weather/file1
    my-dataset           sam        data/my-dataset/weather/file2
    my-dataset           sam        data/my-dataset/weather/file3

Displayed results are sorted based on the value of the first column.

You can specify output formats by passing ``--format`` with a value of ``tabular``,
``json-ld`` or ``json``.

Sometimes you want to filter the files. For this we use ``--dataset``,
``--include`` and ``--exclude`` flags:

.. code-block:: console

    $ renku dataset ls-files --include "file*" --exclude "file3"
    DATASET NAME        ADDED                PATH                           LFS
    ------------------- -------------------  -----------------------------  ----
    my-dataset          2020-02-28 16:49:02  data/my-dataset/weather/file1  *
    my-dataset          2020-02-28 16:49:02  data/my-dataset/weather/file2  *

Dataset files can be listed for a specific version (tag) of a dataset using the
``--tag`` option. In this case, files from datasets which have that specific
tag are displayed.


Unlink a file from a dataset:

.. code-block:: console

    $ renku dataset unlink my-dataset --include file1
    OK

.. cheatsheet::
   :group: Working with Renku Datasets
   :command: $ renku dataset unlink <dataset> [--include <path|pattern>]
   :description: Remove files from a dataset.
   :target: rp

Unlink all files within a directory from a dataset:

.. code-block:: console

    $ renku dataset unlink my-dataset --include "weather/*"
    OK

Unlink all files from a dataset:

.. code-block:: console

    $ renku dataset unlink my-dataset
    Warning: You are about to remove following from "my-dataset" dataset.
    .../my-dataset/weather/file1
    .../my-dataset/weather/file2
    .../my-dataset/weather/file3
    Do you wish to continue? [y/N]:

.. note:: The ``unlink`` command does not delete files,
    only the dataset record.
"""

import json
import os
from pathlib import Path
from typing import List, Union, cast

import click
from lazy_object_proxy import Proxy

import renku.ui.cli.utils.color as color
from renku.command.format.dataset_files import DATASET_FILES_COLUMNS, DATASET_FILES_FORMATS
from renku.command.format.dataset_tags import DATASET_TAGS_FORMATS
from renku.command.format.datasets import DATASETS_COLUMNS, DATASETS_FORMATS
from renku.domain_model.constant import NO_VALUE, NoValueType
from renku.ui.cli.utils.click import shell_complete_datasets


@click.group()
def dataset():
    """Dataset commands."""


@dataset.command("ls")
@click.option(
    "--format", type=click.Choice(list(DATASETS_FORMATS.keys())), default="tabular", help="Choose an output format."
)
@click.option(
    "-c",
    "--columns",
    type=click.STRING,
    default="id,name,title,version,datadir,storage",
    metavar="<columns>",
    help="Comma-separated list of column to display: {}.".format(", ".join(DATASETS_COLUMNS.keys())),
    show_default=True,
)
def list_dataset(format, columns):
    """List datasets."""
    from renku.command.dataset import list_datasets_command

    if format not in DATASETS_FORMATS:
        raise click.BadParameter(f"format '{format}' not supported")

    result = list_datasets_command().lock_dataset().build().execute()

    click.echo(DATASETS_FORMATS[format](result.output, columns=columns))


@dataset.command()
@click.argument("name")
@click.option("-t", "--title", default=None, type=click.STRING, help="Title of the dataset.")
@click.option("-d", "--description", default=None, type=click.STRING, help="Dataset's description.")
@click.option(
    "-c",
    "--creator",
    "creators",
    default=None,
    multiple=True,
    help="Creator's name, email, and affiliation. Accepted format is 'Forename Surname <email> [affiliation]'.",
)
@click.option(
    "-m",
    "--metadata",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Custom metadata to be associated with the dataset.",
)
@click.option("-k", "--keyword", default=None, multiple=True, type=click.STRING, help="List of keywords.")
@click.option("-s", "--storage", default=None, type=click.STRING, help="URI of the cloud storage backend.")
@click.option(
    "--datadir",
    default=None,
    type=click.Path(),
    help="Dataset's data directory (defaults to 'data/<dataset name>').",
)
def create(name, title, description, creators, metadata, keyword, storage, datadir):
    """Create an empty dataset in the current repo."""
    from renku.command.dataset import create_dataset_command
    from renku.core.util.metadata import construct_creators
    from renku.ui.cli.utils.callback import ClickCallback

    communicator = ClickCallback()
    creators = creators or []

    custom_metadata = None

    if metadata:
        custom_metadata = json.loads(Path(metadata).read_text())

    if creators:
        creators, _ = construct_creators(creators)

    result = (
        create_dataset_command()
        .with_communicator(communicator)
        .build()
        .execute(
            name=name,
            title=title,
            description=description,
            creators=creators,
            keywords=keyword,
            custom_metadata=custom_metadata,
            storage=storage,
            datadir=datadir,
        )
    )

    new_dataset = result.output

    click.echo(f'Use the name "{new_dataset.name}" to refer to this dataset.')
    click.secho("OK", fg=color.GREEN)


@dataset.command()
@click.argument("name", shell_complete=shell_complete_datasets)
@click.option("-t", "--title", default=NO_VALUE, type=click.UNPROCESSED, help="Title of the dataset.")
@click.option("-d", "--description", default=NO_VALUE, type=click.UNPROCESSED, help="Dataset's description.")
@click.option(
    "-c",
    "--creator",
    "creators",
    default=[NO_VALUE],
    multiple=True,
    type=click.UNPROCESSED,
    help="Creator's name, email, and affiliation. Accepted format is 'Forename Surname <email> [affiliation]'.",
)
@click.option(
    "-m",
    "--metadata",
    default=NO_VALUE,
    type=click.UNPROCESSED,
    help="Custom metadata to be associated with the dataset.",
)
@click.option(
    "--metadata-source",
    default=NO_VALUE,
    type=click.UNPROCESSED,
    help="Set the source field in the metadata when editing it if not provided, then the default is 'renku'.",
)
@click.option(
    "-k",
    "--keyword",
    "keywords",
    default=[NO_VALUE],
    multiple=True,
    type=click.UNPROCESSED,
    help="List of keywords or tags.",
)
@click.option(
    "-u",
    "--unset",
    default=[],
    multiple=True,
    type=click.Choice(["keywords", "k", "images", "i", "metadata", "m"]),
    help="Remove keywords from dataset.",
)
def edit(name, title, description, creators, metadata, metadata_source, keywords, unset):
    """Edit dataset metadata."""
    from renku.command.dataset import edit_dataset_command
    from renku.core.util.metadata import construct_creators
    from renku.ui.cli.utils.callback import ClickCallback

    images: Union[None, NoValueType] = NO_VALUE

    if list(creators) == [NO_VALUE]:
        creators = NO_VALUE

    if list(keywords) == [NO_VALUE]:
        keywords = NO_VALUE

    if "k" in unset or "keywords" in unset:
        if keywords is not NO_VALUE:
            raise click.UsageError("Cant use '--keyword' together with unsetting keyword")
        keywords = None

    if "m" in unset or "metadata" in unset:
        if metadata is not NO_VALUE:
            raise click.UsageError("Cant use '--metadata' together with unsetting metadata")
        metadata = None

    if "i" in unset or "images" in unset:
        images = None

    if metadata_source is not NO_VALUE and metadata is NO_VALUE:
        raise click.UsageError("The '--metadata-source' option can only be used with the '--metadata' flag")

    if metadata_source is NO_VALUE and metadata is not NO_VALUE:
        metadata_source = "renku"

    custom_metadata = metadata
    no_email_warnings = False

    if creators and creators is not NO_VALUE:
        creators, no_email_warnings = construct_creators(creators, ignore_email=True)  # type: ignore

    if metadata and metadata is not NO_VALUE:
        path = Path(metadata)

        if not path.exists():
            raise click.UsageError(f"Path {path} does not exist.")
        custom_metadata = json.loads(Path(metadata).read_text())

    updated = (
        edit_dataset_command()
        .build()
        .execute(
            name=name,
            title=title,
            description=description,
            creators=creators,
            keywords=keywords,
            images=images,
            custom_metadata=custom_metadata,
            custom_metadata_source=metadata_source,
        )
    ).output

    if not updated:
        click.echo(
            "Nothing to update. "
            "Check available fields with `renku dataset edit --help`\n\n"
            'Hint: `renku dataset edit --title "new title"`'
        )
    else:
        click.echo("Successfully updated: {}.".format(", ".join(updated.keys())))
        if no_email_warnings:
            click.echo(
                ClickCallback.WARNING + "No email or wrong format for: " + ", ".join(cast(List[str], no_email_warnings))
            )


@dataset.command("show")
@click.option("-t", "--tag", default=None, type=click.STRING, help="Tag for which to show dataset metadata.")
@click.argument("name", shell_complete=shell_complete_datasets)
def show(tag, name):
    """Show metadata of a dataset."""
    from renku.command.dataset import show_dataset_command
    from renku.ui.cli.utils.terminal import print_markdown

    result = show_dataset_command().build().execute(name=name, tag=tag)
    ds = result.output

    click.echo(click.style("Name: ", bold=True, fg=color.MAGENTA) + click.style(ds["name"], bold=True))
    click.echo(click.style("Created: ", bold=True, fg=color.MAGENTA) + (ds.get("created_at", "") or ""))
    click.echo(click.style("Data Directory: ", bold=True, fg=color.MAGENTA) + str(ds.get("data_directory", "") or ""))

    creators = []
    for creator in ds.get("creators", []):
        if creator["affiliation"]:
            creators.append(f"{creator['name']} <{creator['email']}> [{creator['affiliation']}]")
        else:
            creators.append(f"{creator['name']} <{creator['email']}>")

    click.echo(click.style("Creator(s): ", bold=True, fg=color.MAGENTA) + ", ".join(creators))
    click.echo(click.style("Keywords: ", bold=True, fg=color.MAGENTA) + ", ".join(ds.get("keywords") or []))

    click.echo(click.style("Version: ", bold=True, fg=color.MAGENTA) + (ds.get("version") or ""))

    click.echo(click.style("Storage: ", bold=True, fg=color.MAGENTA) + (ds.get("storage") or ""))

    click.echo(click.style("Annotations: ", bold=True, fg=color.MAGENTA))
    if ds["annotations"]:
        click.echo(json.dumps(ds.get("annotations", ""), indent=2))

    click.echo(click.style("Title: ", bold=True, fg=color.MAGENTA) + click.style(ds.get("title", ""), bold=True))

    click.echo(click.style("Description: ", bold=True, fg=color.MAGENTA))
    print_markdown(ds.get("description") or "")


def add_provider_options(*param_decls, **attrs):
    """Sets dataset export provider option groups on the dataset add command."""
    from renku.core.dataset.providers.factory import ProviderFactory
    from renku.ui.cli.utils.click import create_options

    providers = [p for p in ProviderFactory.get_add_providers() if p.get_add_parameters()]  # type: ignore
    return create_options(providers=providers, parameter_function="get_add_parameters")


@dataset.command()
@click.argument("name", shell_complete=shell_complete_datasets)
@click.argument("urls", type=click.Path(), nargs=-1)
@click.option("-f", "--force", is_flag=True, help="Allow adding otherwise ignored files.")
@click.option("-o", "--overwrite", is_flag=True, help="Overwrite existing files.")
@click.option("-c", "--create", is_flag=True, help="Create dataset if it does not exist.")
@click.option("-d", "--destination", default="", help="Destination directory within the dataset path")
@click.option("--storage", default=None, type=click.STRING, help="URI of the cloud storage backend.")
@click.option(
    "--datadir",
    default=None,
    type=click.Path(),
    help="Dataset's data directory (defaults to 'data/<dataset name>').",
)
@add_provider_options()
def add(name, urls, force, overwrite, create, destination, datadir, **kwargs):
    """Add data to a dataset."""
    from renku.command.dataset import add_to_dataset_command
    from renku.ui.cli.utils.callback import ClickCallback

    communicator = ClickCallback()
    result = (
        add_to_dataset_command()
        .with_communicator(communicator)
        .build()
        .execute(
            urls=list(urls),
            dataset_name=name,
            force=force,
            overwrite=overwrite,
            create=create,
            destination=destination,
            datadir=datadir,
            **kwargs,
        )
    )

    dataset = result.output
    if dataset.storage:
        communicator.info(f"To download files from the remote storage use 'renku dataset pull {dataset.name}'")

    click.secho("OK", fg=color.GREEN)


@dataset.command("ls-files")
@click.argument("names", nargs=-1, shell_complete=shell_complete_datasets)
@click.option("-t", "--tag", default=None, type=click.STRING, help="Tag for which to show dataset files.")
@click.option(
    "--creators",
    help="Filter files which where authored by specific creators. Multiple creators are specified by comma.",
)
@click.option("-I", "--include", default=None, multiple=True, help="Include files matching given pattern.")
@click.option("-X", "--exclude", default=None, multiple=True, help="Exclude files matching given pattern.")
@click.option(
    "--format",
    type=click.Choice(list(DATASET_FILES_FORMATS.keys())),
    default="tabular",
    help="Choose an output format.",
)
@click.option(
    "-c",
    "--columns",
    type=click.STRING,
    default="dataset_name,path,size,added,lfs",
    metavar="<columns>",
    help="Comma-separated list of column to display: {}.".format(", ".join(DATASET_FILES_COLUMNS.keys())),
    show_default=True,
)
def ls_files(names, tag, creators, include, exclude, format, columns):
    """List files in dataset."""
    from renku.command.dataset import list_files_command

    if format not in DATASETS_FORMATS:
        raise click.BadParameter(f"Format '{format}' not supported")

    result = (
        list_files_command()
        .lock_dataset()
        .build()
        .execute(datasets=names, tag=tag, creators=creators, include=include, exclude=exclude)
    )

    click.echo(DATASET_FILES_FORMATS[format](result.output, columns=columns))


@dataset.command()
@click.argument("name", shell_complete=shell_complete_datasets)
@click.option("-I", "--include", multiple=True, help="Include files matching given pattern.")
@click.option("-X", "--exclude", multiple=True, help="Exclude files matching given pattern.")
@click.option("-y", "--yes", is_flag=True, help="Confirm unlinking of all files.")
def unlink(name, include, exclude, yes):
    """Remove matching files from a dataset."""
    from renku.command.dataset import file_unlink_command
    from renku.core import errors
    from renku.ui.cli.utils.callback import ClickCallback

    if not include and not exclude:
        raise errors.ParameterError(
            "include or exclude filters not found.\n"
            "Check available filters with 'renku dataset unlink --help'\n"
            "Hint: 'renku dataset unlink my-dataset -I path'"
        )

    communicator = ClickCallback()
    file_unlink_command().with_communicator(communicator).build().execute(
        name=name, include=include, exclude=exclude, yes=yes
    )
    click.secho("OK", fg=color.GREEN)


@dataset.command("rm")
@click.argument("name")
def remove(name):
    """Delete a dataset."""
    from renku.command.dataset import remove_dataset_command

    remove_dataset_command().build().execute(name)
    click.secho("OK", fg=color.GREEN)


@dataset.command("tag")
@click.argument("name", shell_complete=shell_complete_datasets)
@click.argument("tag")
@click.option("-d", "--description", default="", help="A description for this tag")
@click.option("-f", "--force", is_flag=True, help="Allow overwriting existing tags.")
def tag(name, tag, description, force):
    """Create a tag for a dataset."""
    from renku.command.dataset import add_dataset_tag_command

    add_dataset_tag_command().build().execute(dataset_name=name, tag=tag, description=description, force=force)
    click.secho("OK", fg=color.GREEN)


@dataset.command("rm-tags")
@click.argument("name", shell_complete=shell_complete_datasets)
@click.argument("tags", nargs=-1)
def remove_tags(name, tags):
    """Remove tags from a dataset."""
    from renku.command.dataset import remove_dataset_tags_command

    remove_dataset_tags_command().build().execute(dataset_name=name, tags=tags)
    click.secho("OK", fg=color.GREEN)


@dataset.command("ls-tags")
@click.argument("name", shell_complete=shell_complete_datasets)
@click.option(
    "--format", type=click.Choice(list(DATASET_TAGS_FORMATS.keys())), default="tabular", help="Choose an output format."
)
def ls_tags(name, format):
    """List all tags of a dataset."""
    from renku.command.dataset import list_tags_command

    result = list_tags_command().lock_dataset().build().execute(dataset_name=name, format=format)
    click.echo(result.output)


def export_provider_argument(*param_decls, **attrs):
    """Sets dataset export provider argument on the dataset export command."""

    def wrapper(f):
        from click import argument

        def get_providers_names():
            from renku.core.dataset.providers.factory import ProviderFactory

            return [p.name.lower() for p in ProviderFactory.get_export_providers()]  # type: ignore

        return argument("provider", type=click.Choice(Proxy(get_providers_names)))(f)

    return wrapper


def export_provider_options(*param_decls, **attrs):
    """Sets dataset export provider option groups on the dataset export command."""
    from renku.core.dataset.providers.factory import ProviderFactory
    from renku.ui.cli.utils.click import create_options

    providers = [p for p in ProviderFactory.get_export_providers() if p.get_export_parameters()]  # type: ignore
    return create_options(providers=providers, parameter_function="get_export_parameters")


@dataset.command()
@click.argument("name", shell_complete=shell_complete_datasets)
@export_provider_argument()
@click.option("-t", "--tag", help="Dataset tag to export")
@export_provider_options()
def export(name, provider, tag, **kwargs):
    """Export data to 3rd party provider."""
    from renku.command.dataset import export_dataset_command
    from renku.core import errors
    from renku.ui.cli.utils.callback import ClickCallback

    try:
        communicator = ClickCallback()
        export_dataset_command().lock_dataset().with_communicator(communicator).build().execute(
            name=name, provider_name=provider, tag=tag, **kwargs
        )
    except (ValueError, errors.InvalidAccessToken, errors.DatasetNotFound, errors.RequestError) as e:
        raise click.BadParameter(str(e))

    click.secho("OK", fg=color.GREEN)


def import_provider_options(*param_decls, **attrs):
    """Sets dataset import provider option groups on the dataset import command."""
    from renku.core.dataset.providers.factory import ProviderFactory
    from renku.ui.cli.utils.click import create_options

    providers = [p for p in ProviderFactory.get_import_providers() if p.get_import_parameters()]  # type: ignore
    return create_options(providers=providers, parameter_function="get_import_parameters")


@dataset.command("import")
@click.argument("uri")
@click.option("--short-name", "--name", "name", default=None, help="A convenient name for dataset.")
@click.option("-x", "--extract", is_flag=True, help="Extract files before importing to dataset.")
@click.option("-y", "--yes", is_flag=True, help="Bypass download confirmation.")
@click.option(
    "--datadir",
    default=None,
    type=click.Path(),
    help="Dataset's data directory (defaults to 'data/<dataset name>').",
)
@import_provider_options()
def import_(uri, name, extract, yes, datadir, **kwargs):
    """Import data from a 3rd party provider or another renku project.

    Supported providers: [Dataverse, Renku, Zenodo]
    """
    from renku.command.dataset import import_dataset_command
    from renku.ui.cli.utils.callback import ClickCallback

    communicator = ClickCallback()
    import_dataset_command().with_communicator(communicator).build().execute(
        uri=uri, name=name, extract=extract, yes=yes, datadir=datadir, **kwargs
    )

    click.secho(" " * 79 + "\r", nl=False)
    click.secho("OK", fg=color.GREEN)


@dataset.command()
@click.pass_context
@click.argument("names", nargs=-1, shell_complete=shell_complete_datasets)
@click.option(
    "--creators",
    help="Filter files which where authored by specific creators. Multiple creators are specified by comma.",
)
@click.option("-I", "--include", default=None, multiple=True, help="Include files matching given pattern.")
@click.option("-X", "--exclude", default=None, multiple=True, help="Exclude files matching given pattern.")
@click.option("--ref", default=None, help="Update to a specific commit/tag/branch.")
@click.option("--delete", is_flag=True, help="Delete local files that are deleted from remote.")
@click.option("-e", "--external", is_flag=True, help="Deprecated")
@click.option("--no-external", is_flag=True, help="Skip updating external data.")
@click.option("--no-local", is_flag=True, help="Skip updating local files.")
@click.option("--no-remote", is_flag=True, help="Skip updating remote files.")
@click.option("-c", "--check-data-directory", is_flag=True, help="Check datasets' data directories for new files.")
@click.option("--all", "-a", "update_all", is_flag=True, default=False, help="Update all datasets.")
@click.option("-n", "--dry-run", is_flag=True, help="Show what would have been changed")
@click.option(
    "--plain",
    is_flag=True,
    default=False,
    help="Show result as one entry per line for machine readability."
    " 'd' = dataset update, 'f' = file update, 'r' = file removed.",
)
def update(
    ctx,
    names,
    creators,
    include,
    exclude,
    ref,
    delete,
    external,
    no_external,
    no_local,
    no_remote,
    check_data_directory,
    update_all,
    dry_run,
    plain,
):
    """Updates files in dataset from a remote Git repo."""
    from renku.command.dataset import update_datasets_command
    from renku.core import errors
    from renku.ui.cli.utils.callback import ClickCallback

    communicator = ClickCallback()

    if external and no_external:
        raise errors.ParameterError("Cannot pass both '--external' and '--no-external'")
    elif external:
        communicator.warn("'-e/--external' argument is deprecated")

    if not update_all and not names and not include and not exclude and not dry_run:
        raise errors.ParameterError("Either NAMES, -a/--all, -n/--dry-run, or --include/--exclude should be specified")

    if names and update_all:
        raise errors.ParameterError("Cannot pass dataset names with -a/--all")
    elif (include or exclude) and update_all:
        raise errors.ParameterError("Cannot pass --include/--exclude with -a/--all")

    result = (
        update_datasets_command(dry_run=dry_run)
        .with_communicator(communicator)
        .build()
        .execute(
            names=list(names),
            creators=creators,
            include=include,
            exclude=exclude,
            ref=ref,
            delete=delete,
            no_local=no_local,
            no_remote=no_remote,
            check_data_directory=check_data_directory,
            update_all=update_all,
            dry_run=dry_run,
            plain=plain,
        )
    )

    if dry_run:
        datasets, dataset_files = result.output

        def get_dataset_files(records):
            from renku.command.format.tabulate import tabulate

            columns = {"path": ("path", None), "dataset": ("dataset.name", "dataset"), "external": ("external", None)}
            return tabulate(collection=records, columns="path,dataset,external", columns_mapping=columns)

        if not datasets and not dataset_files:
            if not plain:
                click.secho("Everything is up-to-date", fg=color.GREEN)
            return

        if datasets:
            ds_names = sorted([d.name for d in datasets])
            if plain:
                ds_names = [f"d {n}" for n in ds_names]
                click.echo(os.linesep.join(ds_names) + os.linesep)
            else:
                names = "\n\t".join(ds_names)
                click.echo(f"The following imported datasets will be updated:\n\t{names}\n")

        if not dataset_files:
            return

        files = [f for f in dataset_files if not f.deleted]
        if files:
            if plain:
                files = [f"f {f.path} {f.dataset.name}" for f in files]
                click.echo(os.linesep.join(files) + os.linesep)
            else:
                files = get_dataset_files(files)
                click.echo(f"The following files will be updated:\n\n{files}\n")

        deleted_files = [f for f in dataset_files if f.deleted]
        if deleted_files:
            if plain:
                files = [f"r {f.path} {f.dataset.name}" for f in deleted_files]
                click.echo(os.linesep.join(files))
            else:
                files = get_dataset_files(deleted_files)
                message = " (pass '--delete' to remove them from datasets' metadata)" if not delete else ""
                click.echo(f"The following files will be deleted{message}:\n\n{files}\n")

        ctx.exit(1)


@dataset.command()
@click.argument("name", shell_complete=shell_complete_datasets)
@click.option(
    "-l",
    "--location",
    default=None,
    type=click.Path(exists=False, file_okay=False, writable=True),
    help="A directory to copy data to, instead of the dataset's data directory.",
)
def pull(name, location):
    """Pull data from a cloud storage."""
    from renku.command.dataset import pull_cloud_storage_command
    from renku.ui.cli.utils.callback import ClickCallback

    communicator = ClickCallback()
    pull_cloud_storage_command().with_communicator(communicator).build().execute(name=name, location=location)


@dataset.command()
@click.argument("name", shell_complete=shell_complete_datasets)
@click.option(
    "-e",
    "--existing",
    default=None,
    type=click.Path(exists=True, file_okay=False),
    help="Use an existing mount point instead of mounting the remote storage.",
)
@click.option("-u", "--unmount", is_flag=True, help="Unmount dataset's backend storage.")
@click.option("-y", "--yes", is_flag=True, help="No prompt when removing non-empty dataset's data directory.")
def mount(name, existing, unmount, yes):
    """Mount a cloud storage in the dataset's data directory."""
    from renku.command.dataset import mount_cloud_storage_command
    from renku.ui.cli.utils.callback import ClickCallback

    command = mount_cloud_storage_command(unmount=unmount).with_communicator(ClickCallback()).build()

    if unmount:
        command.execute(name=name)
    else:
        command.execute(name=name, existing=existing, yes=yes)


@dataset.command()
@click.argument("name", shell_complete=shell_complete_datasets)
def unmount(name):
    """Unmount a backend storage in the dataset's data directory."""
    from renku.command.dataset import unmount_cloud_storage_command
    from renku.ui.cli.utils.callback import ClickCallback

    unmount_cloud_storage_command().with_communicator(ClickCallback()).build().execute(name=name)



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku workflow commands."""


from renku.command.command_builder.command import Command
from renku.core.workflow.activity import revert_activity
from renku.core.workflow.execute import execute_workflow, iterate_workflow
from renku.core.workflow.plan import (
    compose_workflow,
    edit_workflow,
    export_workflow,
    list_workflows,
    remove_plan,
    search_workflows,
    show_workflow,
    visualize_graph,
    workflow_inputs,
    workflow_outputs,
)


def search_workflows_command():
    """Command to get all the workflows whose Plan.name are greater than or equal to the given name."""
    return Command().command(search_workflows).require_migration().with_database(write=False)


def list_workflows_command():
    """Command to list or manage workflows with subcommands."""
    return Command().command(list_workflows).require_migration().with_database(write=False)


def remove_plan_command():
    """Command that removes the workflow named <name>."""
    return Command().command(remove_plan).require_clean().with_database(write=True).with_commit()


def show_workflow_command():
    """Command that the details of a workflow."""
    return Command().command(show_workflow).require_migration().with_database(write=False)


def compose_workflow_command():
    """Command that creates a composite of several workflows."""
    return (
        Command().command(compose_workflow).require_migration().require_clean().with_database(write=True).with_commit()
    )


def edit_workflow_command():
    """Command that edits the properties of a given workflow."""
    return Command().command(edit_workflow).require_clean().with_database(write=True).with_commit()


def export_workflow_command():
    """Command that exports a workflow into a given format."""
    return Command().command(export_workflow).require_migration().with_database(write=False)


def workflow_inputs_command():
    """Command that shows inputs used by workflows."""
    return Command().command(workflow_inputs).require_migration().with_database(write=False)


def workflow_outputs_command():
    """Command that shows inputs used by workflows."""
    return Command().command(workflow_outputs).require_migration().with_database(write=False)


def execute_workflow_command(skip_metadata_update: bool):
    """Command that executes a workflow."""
    command = Command().command(execute_workflow).require_migration()
    if skip_metadata_update:
        command = command.with_database(write=False)
    else:
        command = command.with_database(write=True).with_commit()
    return command


def visualize_graph_command():
    """Execute the graph visualization command."""
    return Command().command(visualize_graph).require_migration().with_database(write=False)


def iterate_workflow_command(skip_metadata_update: bool):
    """Command that executes several workflows given a set of variables."""
    command = Command().command(iterate_workflow).require_migration().require_clean()
    if skip_metadata_update:
        command = command.with_database(write=False)
    else:
        command = command.with_database(write=True).with_commit()
    return command


def revert_activity_command():
    """Command that reverts an activity."""
    return (
        Command().command(revert_activity).require_migration().require_clean().with_database(write=True).with_commit()
    )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project context."""

import contextlib
import threading
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Generator, List, NamedTuple, Optional, Union

from renku.core import errors
from renku.core.constant import (
    APP_NAME,
    CONFIG_NAME,
    DATA_DIR_CONFIG_KEY,
    DATABASE_PATH,
    DATASET_IMAGES,
    DEFAULT_DATA_DIR,
    DOCKERFILE,
    LOCK_SUFFIX,
    POINTERS,
    RENKU_HOME,
    TEMPLATE_CHECKSUMS,
)

if TYPE_CHECKING:
    from renku.domain_model.project import Project
    from renku.infrastructure.database import Database
    from renku.infrastructure.repository import Remote, Repository


class ProjectContext(threading.local):
    """A configuration class to hold global configuration."""

    external_storage_requested = True
    """External storage (e.g. LFS) requested for Renku command."""

    def __init__(self) -> None:
        self._context_stack: List[ProjectProperties] = []

    def __del__(self):
        self.clear()

    @property
    def database(self) -> "Database":
        """Current database."""
        if not self._top.database:
            from renku.infrastructure.database import Database

            self._top.database = Database.from_path(self.database_path)

        return self._top.database

    @property
    def database_path(self) -> Path:
        """Path to the metadata storage directory."""
        return self.metadata_path / DATABASE_PATH

    @property
    def datadir(self) -> str:
        """Define a name of the folder for storing datasets."""
        from renku.core.config import get_value
        from renku.domain_model.enums import ConfigFilter

        if not self._top.datadir:
            datadir = get_value("renku", DATA_DIR_CONFIG_KEY, config_filter=ConfigFilter.LOCAL_ONLY)
            self._top.datadir = datadir or DEFAULT_DATA_DIR

        return self._top.datadir

    @datadir.setter
    def datadir(self, value: str):
        """Set the current datadir."""
        self._top.datadir = value

    @property
    def dataset_images_path(self) -> Path:
        """Return a ``Path`` instance of Renku dataset metadata folder."""
        return self.path / RENKU_HOME / DATASET_IMAGES

    @property
    def dockerfile_path(self) -> Path:
        """Path to the Dockerfile."""
        return self.path / DOCKERFILE

    @property
    def ssh_authorized_keys_path(self) -> Path:
        """Path to SSH authorized keys."""
        return self.path / ".ssh" / "authorized_keys"

    @property
    def global_config_dir(self) -> str:
        """Return user's config directory."""
        import click

        return click.get_app_dir(APP_NAME, force_posix=True)

    @property
    def global_config_path(self) -> Path:
        """Renku global (user's) config path."""
        config = Path(self.global_config_dir)
        if not config.exists():
            config.mkdir(parents=True)

        return config / CONFIG_NAME

    @property
    def latest_agent(self) -> Optional[str]:
        """Returns latest agent version used in the repository."""
        try:
            return self.project.agent_version
        except ValueError as e:
            if "Project name not set" in str(e):
                return None
            raise

    @property
    def local_config_path(self) -> Path:
        """Renku local (project) config path."""
        return self.metadata_path / CONFIG_NAME

    @property
    def lock(self):
        """Create a Renku config lock."""
        from renku.core.util.contexts import Lock

        return Lock(filename=self.metadata_path.with_suffix(LOCK_SUFFIX), mode="exclusive")

    @property
    def metadata_path(self) -> Path:
        """Current project's metadata (RENKU_HOME) path."""
        return self.path / RENKU_HOME

    @property
    def path(self) -> Path:
        """Current project path."""
        return self._top.path

    @property
    def pointers_path(self) -> Path:
        """Return a ``Path`` instance of Renku pointer files folder."""
        path = self.path / RENKU_HOME / POINTERS
        path.mkdir(exist_ok=True)
        return path

    @property
    def project(self) -> "Project":
        """Return the Project instance."""
        from renku.command.command_builder.command import inject
        from renku.core.interface.project_gateway import IProjectGateway

        project_gateway = inject.instance(IProjectGateway)
        # NOTE: Don't cache the project since it can be updated in the ``ProjectGateway``
        return project_gateway.get_project()

    @property
    def remote(self) -> "ProjectRemote":
        """Return host, owner and name of the remote if it exists."""
        from renku.core.util.git import get_remote

        repository = self.repository

        remote = get_remote(repository=repository)

        if not remote and len(repository.remotes) > 1:
            remote = repository.remotes.get("origin")

        return ProjectRemote.from_remote(remote=remote)

    @property
    def repository(self) -> "Repository":
        """Return current context's repository."""
        if not self._top.repository:
            from renku.infrastructure.repository import Repository

            try:
                self._top.repository = Repository(project_context.path)
            except errors.GitError as e:
                raise ValueError from e

        return self._top.repository

    @repository.setter
    def repository(self, value: Optional["Repository"]):
        """Set the current repository."""
        self._top.repository = value

    @property
    def template_checksums_path(self):
        """Return a ``Path`` instance to the template checksums file."""
        return self.metadata_path / TEMPLATE_CHECKSUMS

    @property
    def transaction_id(self) -> str:
        """Get a transaction id for the current context to be used for grouping git commits."""
        if not self._top.transaction_id:
            self._top.transaction_id = uuid.uuid4().hex

        return f"\n\nrenku-transaction: {self._top.transaction_id}"

    @property
    def _top(self) -> "ProjectProperties":
        """Return current context."""
        if self._context_stack:
            return self._context_stack[-1]

        raise errors.ProjectContextError("No project context was pushed")

    def has_context(self, path: Optional[Union[Path, str]] = None) -> bool:
        """Return if at least one context which is equal to path (if not None) is pushed."""
        return True if self._context_stack and (path is None or self.path == Path(path).resolve()) else False

    def clear(self) -> None:
        """Remove all contexts and reset the state without committing intermediate changes.

        NOTE: This method should be used only in tests.
        """
        while self._context_stack:
            if self._top.repository:
                self._top.repository.close()
            self._context_stack.pop()

        self.external_storage_requested = True

    def pop_context(self) -> "ProjectProperties":
        """Pop current project context from stack.

        Returns:
            Path: the popped project path.
        """
        if self._context_stack:
            if self._top.save_changes and self._top.database:
                self._top.database.commit()

            if self._top.repository:
                self._top.repository.close()

            return self._context_stack.pop()
        else:
            raise IndexError("No more context to pop.")

    def push_path(self, path: Union[Path, str], save_changes: bool = False) -> None:
        """Push a new project path to the stack.

        Arguments:
            path(Union[Path, str]): The path to push.
            save_changes(bool): Whether to save changes to the database or not.
        """
        path = Path(path).resolve()
        self._context_stack.append(ProjectProperties(path=path, save_changes=save_changes))

    def replace_path(self, path: Union[Path, str]):
        """Replace the current project path with a new one if they are different.

        Arguments:
            path(Union[Path, str]): The path to replace with.
        """
        path = Path(path).resolve()

        if not self._context_stack:
            self.push_path(path)
        elif self._top.path != path:
            self._context_stack[-1] = ProjectProperties(path=path)

    @contextlib.contextmanager
    def with_path(
        self, path: Union[Path, str], save_changes: bool = False
    ) -> Generator["ProjectProperties", None, None]:
        """Temporarily push a new project path to the stack.

        Arguments:
            path(Union[Path, str]): The path to push.
            save_changes(bool): Whether to save changes to the database or not.
        """
        with self.with_rollback():
            self.push_path(path=path, save_changes=save_changes)
            yield self._top

    @contextlib.contextmanager
    def with_rollback(self) -> Generator[None, None, None]:
        """Rollback to the current state.

        NOTE: This won't work correctly if the current context is popped or swapped.
        """
        before_top = self._top if self._context_stack else None

        try:
            yield
        finally:
            could_rollback = False

            while self._context_stack:
                if self._top == before_top:
                    could_rollback = True
                    break

                self.pop_context()

            if not could_rollback and before_top is not None:
                raise errors.ProjectContextError(f"Cannot rollback to {before_top.path}.")


project_context: ProjectContext = ProjectContext()


def has_graph_files() -> bool:
    """Return true if database exists."""
    return project_context.database_path.exists() and any(
        f for f in project_context.database_path.iterdir() if f != project_context.database_path / "root"
    )


@dataclass
class ProjectProperties:
    """Various properties of the current project."""

    path: Path
    database: Optional["Database"] = None
    datadir: Optional[str] = None
    repository: Optional["Repository"] = None
    save_changes: bool = False
    transaction_id: Optional[str] = None


class ProjectRemote(NamedTuple):
    """Information about a project's remote."""

    name: Optional[str]
    owner: Optional[str]
    host: Optional[str]

    @classmethod
    def from_remote(cls, remote: Optional["Remote"]) -> "ProjectRemote":
        """Create an instance from a Repository remote."""
        from renku.domain_model.git import GitURL

        if not remote:
            return ProjectRemote(None, None, None)

        url = GitURL.parse(remote.url)

        # NOTE: Remove gitlab unless running on gitlab.com
        hostname_parts = url.hostname.split(".")
        if len(hostname_parts) > 2 and hostname_parts[0] == "gitlab":
            hostname_parts = hostname_parts[1:]

        return ProjectRemote(name=url.name, owner=url.owner, host=".".join(hostname_parts))

    def __bool__(self):
        return bool(self.name or self.owner or self.host)



#
# Copyright 2018-2023- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku run command."""

from typing import List, Optional

from renku.command.command_builder.command import Command
from renku.core.workflow.run import run_command_line
from renku.core.workflow.workflow_file import run_workflow_file


def run_command_line_command():
    """Run command line."""
    command = Command().command(run_command_line)
    return command.require_migration().require_clean().with_database(write=True).with_commit()


def run_workflow_file_command(no_commit: bool, commit_only: Optional[List[str]]):
    """Run a workflow file."""
    command = Command().command(run_workflow_file).require_migration()
    return (
        command.with_database(write=False)
        if no_commit
        else command.with_database(write=True).with_commit(
            commit_only=commit_only, skip_staging=True, skip_dirty_checks=True
        )
    )



#
# Copyright 2017-2023- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plugin hooks for renku workflow customization."""
from typing import TYPE_CHECKING, List, Type

import pluggy

if TYPE_CHECKING:
    from renku.core.dataset.providers.api import ProviderApi

hookspec = pluggy.HookspecMarker("renku")


@hookspec
def dataset_provider() -> "Type[ProviderApi]":  # type: ignore[empty-body]
    """Plugin Hook for different dataset providers.

    Returns:
        A reference to the provider itself
    """
    pass


def get_supported_dataset_providers() -> "List[Type[ProviderApi]]":
    """Returns the currently available interactive session providers."""
    from renku.core.plugin.pluginmanager import get_plugin_manager

    pm = get_plugin_manager()
    return pm.hook.dataset_provider()



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service file cache serializers."""
import uuid

from marshmallow import fields, post_load

from renku.ui.service.cache.models.file import File, FileChunk
from renku.ui.service.serializers.common import CreationSchema, FileDetailsSchema, MandatoryUserSchema


class FileSchema(FileDetailsSchema, MandatoryUserSchema):
    """Schema for file model."""

    override_existing = fields.Boolean(
        load_default=False,
        metadata={"description": "Overried files. Useful when extracting from archives."},
    )

    @post_load
    def make_file(self, data, **options):
        """Construct file object."""
        return File(**data)


class FileChunkSchema(CreationSchema, MandatoryUserSchema):
    """Schema for file model."""

    chunk_file_id = fields.String(load_default=lambda: uuid.uuid4().hex)
    file_name = fields.String(required=True)

    chunked_id = fields.String(required=True)
    relative_path = fields.String(required=True)

    @post_load
    def make_file(self, data, **options):
        """Construct file object."""
        return FileChunk(**data)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku datetime utilities."""

import re
from datetime import datetime, timezone
from typing import Optional

regex = (
    r"^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12]["
    r"0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|["
    r"+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$"
)
match_iso8601 = re.compile(regex).match


def validate_iso8601(str_val):
    """Check if datetime string is in ISO8601 format."""
    try:
        if match_iso8601(str_val) is not None:
            return True
    except re.error:
        pass
    return False


def parse_date(value):
    """Convert date to datetime."""
    from dateutil.parser import parse as date_util_parse_date

    if value is None:
        return
    if isinstance(value, datetime):
        date = value
    else:
        date = date_util_parse_date(value)

    if not date.tzinfo:
        date = _set_to_local_timezone(date)

    return date


def fix_datetime(value) -> Optional[datetime]:
    """Fix timezone of non-aware datetime objects and remove microseconds."""
    if value is None:
        return None

    if isinstance(value, datetime):
        if not value.tzinfo:
            value = _set_to_local_timezone(value)

    return value


def _set_to_local_timezone(value):
    local_tz = local_now().tzinfo
    return value.replace(tzinfo=local_tz)


def local_now(remove_microseconds: bool = False) -> datetime:
    """Return current datetime in local timezone."""
    now = datetime.now(timezone.utc).astimezone()
    return now.replace(microsecond=0) if remove_microseconds else now



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Git utility functions."""

import contextlib
import os
import posixpath
import re
import shutil
import sys
import time
import urllib
from functools import reduce
from pathlib import Path
from subprocess import PIPE, SubprocessError, run
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union, cast
from uuid import uuid4

from renku.core import errors
from renku.infrastructure.repository import DiffChangeType

if TYPE_CHECKING:
    from renku.domain_model.entity import Collection, Entity
    from renku.domain_model.git import GitURL
    from renku.domain_model.provenance.agent import Person, SoftwareAgent
    from renku.infrastructure.repository import Commit, Remote, Repository

COMMIT_DIFF_STRATEGY = "DIFF"
STARTED_AT = int(time.time() * 1e3)

BRANCH_NAME_LIMIT = 250
CLI_GITLAB_ENDPOINT = "repos"
PROTECTED_BRANCH_PREFIX = "renku/autobranch"
RENKU_BACKUP_PREFIX = "renku-backup"


# TODO: Make sure caching is thread-safe
_entity_cache: Dict[Tuple[Optional[str], str], Union["Entity", "Collection"]] = {}


def run_command(command, *paths, separator=None, **kwargs):
    """Execute command by splitting `paths` to make sure that argument list will be within os limits.

    Args:
        command: A list or tuple containing command and its arguments.
        *paths: Paths to run on.
        separator: Separator for `paths` if they need to be passed as string. (Default value = None)
    Raises:
        errors.GitError: If a Git subcommand failed.
    Returns:
        Result of last invocation.

    """
    from renku.infrastructure.repository import split_paths

    result = None

    for batch in split_paths(*paths):
        if separator:
            batch = [separator.join(batch)]

        try:
            if not isinstance(batch, list):
                batch = list(batch)
            result = run(command + batch, **kwargs)

            if result.returncode != 0:
                break
        except KeyboardInterrupt:
            raise
        except SubprocessError as e:
            raise errors.GitError(f"Cannot run command {command} : {e}")

    return result


def is_valid_git_repository(repository: Optional["Repository"]) -> bool:
    """Return if is a git repository and has a valid HEAD.

    Args:
        repository(Optional[Repository]): The repository to check.

    Returns:
        bool: Whether this is a valid Git repository.

    """
    return repository is not None and repository.head.is_valid()


def get_hook_path(path: Path, name: str) -> Path:
    """Return path to the given named hook in the given repository.

    Args:
        path(Path): The current Git repository's path.
        name(str): The name of the hook.

    Returns:
        Path: Path to the hook.

    """
    return path / ".git" / "hooks" / name


def get_oauth_url(url, gitlab_token):
    """Format URL with a username and password.

    Args:
        url: The URL to format.
        gitlab_token: The Gitlab OAuth2 Token.

    Returns:
        The URL with credentials added.
    """
    parsed_url = urllib.parse.urlparse(url)

    if not parsed_url.netloc:
        raise ValueError(f"Invalid http git url: {url}")

    netloc = f"oauth2:{gitlab_token}@{parsed_url.netloc}"
    return parsed_url._replace(netloc=netloc).geturl()


def get_cache_directory_for_repository(url) -> Path:
    """Return a path to project's cache directory.

    Args:
        url: The repository URL.

    Returns:
        Path: The path of the cache.

    """
    from renku.core.constant import CACHE
    from renku.domain_model.project_context import project_context

    return project_context.metadata_path / CACHE / get_full_repository_path(url)


def parse_git_url(url: Optional[str]) -> "GitURL":
    """Return parsed git url.

    Args:
        url(Optional[str]): The URL to parse.
    Raises:
        errors.InvalidGitURL: If ``url`` is empty.
    Returns:
        GitURL: The parsed GitURL.

    """
    from renku.domain_model.git import GitURL

    if not url:
        raise errors.InvalidGitURL("No URL provided.")

    return GitURL.parse(url)


def have_same_remote(url1, url2) -> bool:
    """Checks if two git urls point to the same remote repo ignoring protocol and credentials.

    Args:
        url1: The first URL.
        url2:The second URL.

    Returns:
        bool: True if both URLs point to the same repository.
    """
    try:
        u1 = parse_git_url(url1)
        u2 = parse_git_url(url2)

        return u1.hostname == u2.hostname and u1.path == u2.path
    except (errors.GitError, AttributeError):
        return False


def get_renku_repo_url(remote_url, deployment_hostname=None, access_token=None):
    """Return a repo url that can be authenticated by renku.

    Args:
        remote_url: The repository URL.
        deployment_hostname: The host name used by this deployment (Default value = None).
        access_token: The OAuth2 access token (Default value = None).
    Returns:
        The Renku repository URL with credentials.
    """
    parsed_remote = parse_git_url(remote_url)
    path = parsed_remote.path.strip("/")
    if path.startswith("gitlab/"):
        path = path.replace("gitlab/", "")
    if not path.startswith(f"{CLI_GITLAB_ENDPOINT}/"):
        path = posixpath.join(CLI_GITLAB_ENDPOINT, path)

    credentials = f"renku:{access_token}@" if access_token else ""
    hostname = deployment_hostname or parsed_remote.hostname

    if hostname.startswith("gitlab."):
        hostname = hostname.replace("gitlab.", "", 1)

    return urllib.parse.urljoin(f"https://{credentials}{hostname}", path)


def create_backup_remote(repository: "Repository", remote_name: str, url: str) -> Tuple[str, bool, Optional["Remote"]]:
    """Create a backup for ``remote_name`` and sets its url to ``url``.

    Args:
        repository(Repository): The current repository.
        remote_name(str): The name of the backup remote.
        url(str): The remote URL.

    Returns:
        Tuple[str, bool, Optional[Remote]]: Tuple of backup remote name, whether it existed already and the created
            remote if successful.
    """
    backup_remote_name = f"{RENKU_BACKUP_PREFIX}-{remote_name}"

    backup_exists = any(backup_remote_name == r.name for r in repository.remotes)
    if backup_exists:
        return backup_remote_name, True, None

    try:
        remote = repository.remotes.add(name=backup_remote_name, url=url)
    except errors.GitCommandError:
        return backup_remote_name, False, None
    else:
        return backup_remote_name, False, remote


def set_git_credential_helper(repository: "Repository", hostname):
    """Set up credential helper for renku git."""
    with repository.get_configuration(writable=True) as config:
        config.set_value("credential", "helper", f"!renku credentials --hostname {hostname}")


def get_full_repository_path(url: Optional[str]) -> str:
    """Extract hostname/path of a git repository from its URL.

    Args:
        url(Optional[str]): The URL.

    Returns:
        The hostname plus path extracted from the URL.
    """
    if url is None:
        return ""

    parsed_url = parse_git_url(url)
    return posixpath.join(parsed_url.hostname, parsed_url.path)  # type:ignore


def get_repository_name(url: str) -> str:
    """Extract name of a git repository from its URL.

    Args:
        url(str): The URL to get the repository name from.

    Returns:
        str: The repository name.
    """
    return Path(get_renku_repo_url(url)).stem


def get_committer_agent(commit: "Commit") -> "SoftwareAgent":
    """Return committer SoftwareAgent.

    Args:
        commit(Commit): The commit to check.

    Returns:
        SoftwareAgent: The agent responsible for the commit.
    """
    from renku.domain_model.provenance.agent import SoftwareAgent

    return SoftwareAgent(id=commit.committer.email, name=commit.committer.name)


def get_git_user(repository: Optional["Repository"]) -> Optional["Person"]:
    """Return git user.

    Args:
        repository(Optional[Repository]): The Git repository.

    Returns:
        Optional[Person]: The person associated with the repository.

    """
    from renku.domain_model.provenance.agent import Person

    if repository is None:
        return None

    user = repository.get_user()
    return Person(name=user.name, email=user.email)


def get_remote(
    repository: Optional["Repository"], *, name: Optional[str] = None, url: Optional[str] = None
) -> Optional["Remote"]:
    """Return repository's remote using its name or url or return default remote if any.

    Args:
        repository(Optional[Repository]): The Git repository.
        name(str, optional): The name of the remote (Default value = None).
        url(str, optional): The remote URL (Default value = None).

    Returns:
        Optional[Remote]: The remote, if found.

    """
    if not repository or len(repository.remotes) == 0:
        return None
    elif name:
        return next((r for r in repository.remotes if r.name == name), None)
    elif url:
        return next((r for r in repository.remotes if r.url == url), None)
    elif len(repository.remotes) == 1:
        return repository.remotes[0]
    elif repository.active_branch and repository.active_branch.remote_branch:
        return repository.active_branch.remote_branch.remote

    return None


def check_global_git_user_is_configured():
    """Check that git user information is configured."""
    from renku.infrastructure.repository import Repository

    try:
        Repository.get_global_user()
    except errors.GitConfigurationError:
        return False
    else:
        return True


def is_path_safe(path: Union[Path, str]) -> bool:
    """Check if the path should be used in output.

    Args:
        path(Union[Path, str]): The path to check.

    Returns:
        bool: True if the path is safe else False.
    """
    path = str(path)

    # Should not be in ignore paths.
    if path in {".gitignore", ".gitattributes"}:
        return False

    # Ignore everything in .renku ...
    if path.startswith(".renku"):
        return False

    return True


def get_entity_from_revision(
    repository: "Repository",
    path: Union[Path, str],
    revision: Optional[str] = None,
    bypass_cache: bool = False,
    checksum: Optional[str] = None,
) -> "Entity":
    """Return an Entity instance from given path and revision.

    Args:
        repository(Repository): The current repository.
        path(Union[Path, str]): The path of the entity.
        revision(str, optional): The revision to check at (Default value = None).
        bypass_cache(bool): Whether to ignore cached entries and get information from disk (Default value = False).
        checksum(str, optional): Pre-calculated checksum for performance reasons, will be calculated if not set.
    Returns:
        Entity: The Entity for the given path and revision.

    """
    from renku.domain_model.constant import NON_EXISTING_ENTITY_CHECKSUM
    from renku.domain_model.entity import Collection, Entity

    def get_directory_members(absolute_path: Path) -> List[Entity]:
        """Return first-level files/directories in a directory."""
        members: List[Entity] = []

        for member in absolute_path.iterdir():
            if member.name == ".gitkeep":
                continue

            member_path = member.relative_to(repository.path)

            assert all(member_path != m.path for m in members)

            entity = get_entity_from_revision(repository, member_path, revision, bypass_cache=bypass_cache)
            # NOTE: If a path is not found at a revision we assume that it didn't exist at that revision
            if entity:
                members.append(entity)

        return members

    global _entity_cache
    key = (revision, str(path))
    cached_entry = _entity_cache.get(key)
    if cached_entry and not bypass_cache:
        return cached_entry

    # NOTE: For untracked directory the hash is None; make sure to stage them first before calling this function.
    if not checksum:
        checksum = repository.get_object_hash(revision=revision, path=path)
    # NOTE: If object was not found at a revision it's either removed or exists in a different revision; keep the
    # entity and use revision as checksum
    if isinstance(revision, str) and revision == "HEAD":
        revision = repository.head.commit.hexsha
    checksum = checksum or revision or NON_EXISTING_ENTITY_CHECKSUM
    id = Entity.generate_id(checksum=checksum, path=path)

    absolute_path = repository.path / path
    if str(path) != "." and absolute_path.is_dir():
        members = get_directory_members(absolute_path)
        entity: Union[Entity, Collection] = Collection(id=id, checksum=checksum, path=path, members=members)
    else:
        entity = Entity(id=id, checksum=checksum, path=path)

    _entity_cache[key] = entity

    return entity


def get_git_path(path: Union[Path, str] = ".") -> Path:
    """Return the repository path."""
    # TODO: Implement this using ``git rev-parse --git-dir``
    try:
        path = get_git_repository(path=path).path
    except ValueError:
        path = Path(path)

    return path.resolve()


def get_git_repository(path: Union[Path, str] = ".") -> "Repository":
    """Get Git repository from the current path or any of its parents.

    Args:
        path: Path to start from (Default value = ".").
    Raises:
        ValueError: If not inside a git repository.
    Returns:
        Git repository
    """
    from renku.infrastructure.repository import Repository

    try:
        return Repository(path, search_parent_directories=True)
    except errors.GitError:
        raise ValueError(f"Cannot find a git repository at '{path}'")


def commit_changes(*paths: Union[Path, str], repository: "Repository", message=None) -> List[str]:
    """Commit paths to the repository.

    Args:
        *paths(Union[Path, str]): The paths to commit.
        repository(Repository): The repository to commit to.
        message: The commit message (Default value = None).
    Raises:
        errors.GitError: If paths couldn't be committed.
    Returns:
        List of paths that were committed.
    """
    if len(paths) == 0:
        return []

    try:
        staged_files = {c.a_path for c in repository.staged_changes} if repository.head.is_valid() else set()
        path_to_save = set(paths) - staged_files
        repository.add(*path_to_save)
        saved_paths = [c.b_path for c in repository.staged_changes]

        if saved_paths:
            if not message:
                # Show saved files in message
                max_line_len = 100
                max_total_len = 100000
                message = "Saved changes to: "
                paths_with_lens = cast(
                    List[Tuple[str, int]],
                    reduce(
                        lambda c, x: c + [(x, c[-1][1] + len(x))],
                        saved_paths,
                        cast(List[Tuple[Optional[str], int]], [(None, len(message))]),
                    )[1:],
                )
                # limit first line to max_len characters
                message += " ".join(p if l < max_line_len else "\n\t" + p for p, l in paths_with_lens)

                if len(message) > max_total_len:
                    message = message[: max_total_len - 3] + "..."

            repository.commit(message)
    except errors.GitCommandError as e:
        raise errors.GitError("Cannot commit changes") from e
    else:
        return saved_paths


def push_changes(repository: "Repository", remote: Optional[str] = None, reset: bool = True) -> str:
    """Push to a remote branch. If the remote branch is protected a new remote branch will be created and pushed to.

    Args:
        repository(Repository): The current repository.
        remote(str, optional): The remote to push to (Default value = None).
        reset(bool, optional): Whether to reset active branch to its upstream branch, used if changes get
            pushed to a temporary branch (Default value = True).
    Raises:
        errors.GitError: If there's no remote or the push fails.
    Returns:
        str: Name of the branch that was pushed to.
    """
    from renku.core.util import communication

    if repository.active_branch is None:
        raise errors.GitError("Cannot push changes when repository is in detached HEAD state.")

    if repository.active_branch.remote_branch:
        ref = repository.active_branch.remote_branch.name
        pushed_branch = ref.split("/")[-1]
    else:
        pushed_branch = repository.active_branch.name

    if remote is not None:
        pushed_remote = get_remote(repository, url=remote) or get_remote(repository, name=remote)
        if not pushed_remote:
            if get_remote(repository, name="origin") is not None:
                pushed_branch = uuid4().hex
                pushed_remote_name = uuid4().hex
            else:
                pushed_remote_name = "origin"
            pushed_remote = repository.remotes.add(name=pushed_remote_name, url=remote)
    else:
        pushed_remote = get_remote(repository)
        if not pushed_remote:
            raise errors.GitRemoteNotFoundError("No remote has been set up for the current branch")

    try:
        # NOTE: Push local changes to remote branch.
        merge_conflict = False
        if len(pushed_remote.references) > 0 and repository.active_branch.remote_branch in pushed_remote.references:
            repository.fetch(pushed_remote)
            try:
                repository.pull(pushed_remote, repository.active_branch)
            except errors.GitCommandError as e:
                # NOTE: Couldn't pull, probably due to conflicts, try a merge.
                # NOTE: the error sadly doesn't tell any details.
                unmerged_blobs = repository.unmerged_blobs.values()
                conflicts = (stage != 0 for blobs in unmerged_blobs for stage, _ in blobs)
                if any(conflicts):
                    merge_conflict = True

                    if communication.confirm(
                        "There were conflicts when updating the local data with remote changes,"
                        " do you want to resolve them (if not, a new remote branch will be created)?",
                        warning=True,
                    ):
                        repository.run_git_command("mergetool", "-g")
                        repository.commit("merging conflict", no_edit=True)
                        merge_conflict = False
                    else:
                        repository.reset(hard=True)
                else:
                    raise errors.GitError("Cannot pull changes from remote") from e

        push_failed = False

        if not merge_conflict:
            # NOTE: Try pushing to remote branch which might fail if the branch is protected
            try:
                repository.push(pushed_remote, f"{repository.active_branch.name}:{pushed_branch}")
            except errors.GitCommandError:
                push_failed = True

        if merge_conflict or push_failed:
            # NOTE: Push to a new remote branch and reset the cache.
            last_short_sha = repository.head.commit.hexsha[0:8]
            old_active_branch = str(repository.active_branch)
            fixed_chars_len = len(PROTECTED_BRANCH_PREFIX) + len(last_short_sha) + 2
            if len(old_active_branch) + fixed_chars_len > BRANCH_NAME_LIMIT:
                old_branch_reference = old_active_branch[0 : (BRANCH_NAME_LIMIT - fixed_chars_len)]
            else:
                old_branch_reference = old_active_branch
            pushed_branch = f"{PROTECTED_BRANCH_PREFIX}/{old_branch_reference}/{last_short_sha}"
            try:
                repository.branches.add(pushed_branch)
                repository.checkout(pushed_branch)
                repository.push(pushed_remote, pushed_branch, set_upstream=True)
            except Exception:
                reset = False  # NOTE: Don't reset the repository if push to the new remote branch failed
                raise
            finally:
                repository.checkout(old_active_branch)
                if reset:
                    repository.reset(reference=repository.active_branch.remote_branch, hard=True)
    except errors.GitCommandError as e:
        raise errors.GitError("Cannot push changes") from e

    return pushed_branch


def clone_renku_repository(
    url: str,
    path: Optional[Union[Path, str]],
    gitlab_token=None,
    deployment_hostname=None,
    depth: Optional[int] = None,
    install_githooks=False,
    install_lfs=True,
    skip_smudge=True,
    recursive=True,
    progress=None,
    config: Optional[dict] = None,
    raise_git_except=False,
    checkout_revision=None,
    use_renku_credentials: bool = False,
    reuse_existing_repository: bool = False,
) -> "Repository":
    """Clone a Renku Repository.

    Args:
        url(str): The Git URL to clone.
        path(Union[Path, str]): The path to clone into.
        gitlab_token: The gitlab OAuth2 token (Default value = None).
        deployment_hostname: The hostname of the current renku deployment (Default value = None).
        depth(Optional[int], optional): The clone depth, number of commits from HEAD (Default value = None).
        install_githooks: Whether to install git hooks (Default value = False).
        install_lfs: Whether to install Git LFS (Default value = True).
        skip_smudge: Whether to pull files from Git LFS (Default value = True).
        recursive: Whether to clone recursively (Default value = True).
        progress: The GitProgress object (Default value = None).
        config(Optional[dict], optional): Set configuration for the project (Default value = None).
        raise_git_except: Whether to raise git exceptions (Default value = False).
        checkout_revision: The revision to check out after clone (Default value = None).
        use_renku_credentials(bool, optional): Whether to use Renku provided credentials (Default value = False).
        reuse_existing_repository(bool, optional): Whether to clone over an existing repository (Default value = False).

    Returns:
        The cloned repository.
    """
    from renku.core.login import has_credentials_for_hostname

    parsed_url = parse_git_url(url)

    clone_options = None
    create_backup = False

    if parsed_url.hostname == "localhost":
        absolute_path = Path(parsed_url.path).resolve()
        git_url = str(absolute_path)
    elif parsed_url.scheme in ["http", "https"] and gitlab_token:
        git_url = get_oauth_url(url, gitlab_token)
    elif (
        parsed_url.scheme in ["http", "https"]
        and use_renku_credentials
        and has_credentials_for_hostname(parsed_url.hostname)  # NOTE: Don't change remote URL if no credentials exist
    ):
        clone_options = [f"--config credential.helper='!renku credentials --hostname {parsed_url.hostname}'"]
        deployment_hostname = deployment_hostname or parsed_url.hostname
        git_url = get_renku_repo_url(url, deployment_hostname=deployment_hostname, access_token=None)
        create_backup = True
    else:
        git_url = url

    repository = clone_repository(
        git_url,
        path=path,
        install_githooks=install_githooks,
        depth=depth,
        clean=reuse_existing_repository,
        install_lfs=install_lfs,
        skip_smudge=skip_smudge,
        recursive=recursive,
        progress=progress,
        config=config,
        raise_git_except=raise_git_except,
        checkout_revision=checkout_revision,
        clone_options=clone_options,
    )

    if create_backup:
        create_backup_remote(repository=repository, remote_name="origin", url=url)
        set_git_credential_helper(
            repository=cast("Repository", repository), hostname=deployment_hostname or parsed_url.hostname
        )

    return repository


def clone_repository(
    url,
    path: Optional[Union[Path, str]] = None,
    install_githooks=True,
    install_lfs=True,
    skip_smudge=True,
    recursive=True,
    depth=None,
    progress=None,
    config: Optional[dict] = None,
    raise_git_except=False,
    checkout_revision=None,
    no_checkout: bool = False,
    clean: bool = False,
    clone_options: Optional[List[str]] = None,
) -> "Repository":
    """Clone a Git repository and install Git hooks and LFS.

    Args:
        url: The Git URL to clone.
        path(Union[Path, str], optional): The path to clone into (Default value = None).
        install_githooks: Whether to install git hooks (Default value = True).
        install_lfs: Whether to install Git LFS (Default value = True).
        skip_smudge: Whether to pull files from Git LFS (Default value = True).
        recursive: Whether to clone recursively (Default value = True).
        depth: The clone depth, number of commits from HEAD (Default value = None).
        progress: The GitProgress object (Default value = None).
        config(Optional[dict], optional): Set configuration for the project (Default value = None).
        raise_git_except: Whether to raise git exceptions (Default value = False).
        checkout_revision: The revision to check out after clone (Default value = None).
        no_checkout(bool, optional): Whether to perform a checkout (Default value = False).
        clean(bool, optional): Whether to require the target folder to be clean (Default value = False).
        clone_options(List[str], optional): Additional clone options (Default value = None).

    Returns:
        The cloned repository.
    """
    from renku.core.githooks import install_githooks as install_githooks_function
    from renku.infrastructure.repository import Repository

    path = Path(path) if path else Path(get_repository_name(url))

    def handle_git_exception():
        """Handle git exceptions."""
        if raise_git_except:
            return

        message = f"Cannot clone repo from {url}"

        if progress:
            lines = progress.other_lines + progress.error_lines
            error = "".join([f"\n\t{line}" for line in lines if line.strip()])
            message += f" - error message:\n {error}"

        raise errors.GitError(message)

    def clean_directory():
        if not clean or not path:
            return
        try:
            shutil.rmtree(path)
        except FileNotFoundError:
            pass
        except PermissionError as e:
            raise errors.InvalidFileOperation(f"Cannot delete files in {path}: Permission denied") from e

    def check_and_reuse_existing_repository() -> Optional["Repository"]:
        if path is None or not cast(Path, path).exists():
            return None

        try:
            repository = Repository(path)
            remote = get_remote(repository, name="origin") or get_remote(repository)

            if remote and have_same_remote(remote.url, url):
                repository.reset(hard=True)
                repository.fetch(all=True, tags=True)
                # NOTE: By default we check out remote repository's HEAD since the local HEAD might not point to
                # the default branch.
                default_checkout_revision = checkout_revision or "origin/HEAD"
                repository.checkout(default_checkout_revision)
                try:
                    repository.pull()
                except errors.GitCommandError:  # NOTE: When ref is not a branch, an error is thrown
                    pass
            else:
                # NOTE: not same remote, so don't reuse
                clean_directory()
                return None
        except errors.GitError:  # NOTE: Not a git repository, remote not found, or checkout failed
            clean_directory()
        else:
            return repository

        return None

    def clone(branch, depth):
        os.environ["GIT_LFS_SKIP_SMUDGE"] = "1" if skip_smudge else "0"

        return Repository.clone_from(
            url,
            cast(Path, path),
            branch=branch,
            recursive=recursive,
            depth=depth,
            no_checkout=no_checkout,
            progress=progress,
            clone_options=clone_options,
        )

    assert config is None or isinstance(config, dict), f"Config should be a dict not '{type(config)}'"

    existing_repository = check_and_reuse_existing_repository()
    if existing_repository is not None:
        return existing_repository

    try:
        # NOTE: Try to clone, assuming checkout_revision is a branch or a tag (if it is set)
        repository = clone(branch=checkout_revision, depth=depth)
    except errors.GitCommandError:
        if not checkout_revision:
            handle_git_exception()
            raise

        # NOTE: clone without branch set, in case checkout_revision was not a branch or a tag but a commit
        try:
            repository = clone(branch=None, depth=None)
        except errors.GitCommandError:
            handle_git_exception()
            raise

    if checkout_revision is not None and not no_checkout:
        try:
            repository.checkout(checkout_revision)
        except errors.GitCommandError:
            raise errors.GitReferenceNotFoundError(
                f"Cannot checkout reference '{checkout_revision}' in repository: {url}"
            )

    if config:
        with repository.get_configuration(writable=True) as config_writer:
            for key, value in config.items():
                try:
                    section, option = key.rsplit(".", maxsplit=1)
                except ValueError:
                    raise errors.GitError(f"Cannot write to config: Invalid config '{key}'")

                config_writer.set_value(section, option, value)

    if install_githooks:
        install_githooks_function(force=True, path=repository.path)

    if install_lfs:
        repository.lfs.install(skip_smudge=skip_smudge)

    return repository


def get_git_progress_instance():
    """Return a GitProgress object."""
    from git.remote import RemoteProgress

    class GitProgress(RemoteProgress):
        """Progress printing for GitPython."""

        def __init__(self):
            """Initialize a Git progress printer."""
            super().__init__()
            self._previous_line_length = 0

        def update(self, op_code, cur_count, max_count=None, message=""):
            """Callback for printing Git operation status."""
            self._clear_line()
            print(self._cur_line, end="\r")
            self._previous_line_length = len(self._cur_line) if self._cur_line else 0
            if (op_code & RemoteProgress.END) != 0:
                print()

        def _clear_line(self):
            print(self._previous_line_length * " ", end="\r")

    return GitProgress()


def get_file_size(repository_path: Path, path: str) -> Optional[int]:
    """Return file size for a file inside a git repository."""
    # NOTE: First try to get file size from Git LFS
    try:
        lfs_run = run(
            ("git", "lfs", "ls-files", "--name-only", "--size"),
            stdout=PIPE,
            cwd=repository_path,
            text=True,
        )
    except SubprocessError:
        pass
    else:
        lfs_output = lfs_run.stdout.split("\n")
        # Example line format: relative/path/to/file (7.9 MB)
        pattern = re.compile(r".*\((.*)\)")
        for line in lfs_output:
            if path not in line:
                continue
            match = pattern.search(line)
            if not match:
                continue
            size_info = match.groups()[0].split()
            if len(size_info) != 2:
                continue
            try:
                size = float(size_info[0])
            except ValueError:
                continue
            unit = size_info[1].strip().lower()
            conversions = {"b": 1, "kb": 1e3, "mb": 1e6, "gb": 1e9}
            multiplier = conversions.get(unit, None)
            if multiplier is None:
                continue
            return int(size * multiplier)

    # Return size of the file on disk
    full_path = repository_path / path
    return os.path.getsize(full_path) if full_path.exists() else None


def shorten_message(message: str, line_length: int = 100, body_length: int = 65000) -> str:
    """Wraps and shortens a commit message.

    Args:
        message(str): message to adjust.
        line_length(int, optional): maximum line length before wrapping. 0 for infinite (Default value = 100).
        body_length(int, optional): maximum body length before cut. 0 for infinite (Default value = 65000).
    Raises:
        ParameterError: If line_length or body_length < 0
    Returns:
        message wrapped and trimmed.

    """
    if line_length < 0:
        raise errors.ParameterError("the length can't be negative.", "line_length")

    if body_length < 0:
        raise errors.ParameterError("the length can't be negative.", "body_length")

    if body_length and len(message) > body_length:
        message = message[: body_length - 3] + "..."

    if line_length == 0 or len(message) <= line_length:
        return message

    lines = message.split(" ")
    lines = [
        line
        if len(line) < line_length
        else "\n\t".join(line[o : o + line_length] for o in range(0, len(line), line_length))
        for line in lines
    ]

    # NOTE: tries to preserve message spacing.
    wrapped_message = reduce(
        lambda c, x: (f"{c[0]} {x}", c[1] + len(x) + 1)
        if c[1] + len(x) <= line_length
        else (f"{c[0]}\n\t" + x, len(x)),
        lines,
        ("", 0),
    )[0]
    return wrapped_message[1:]


def get_in_submodules(
    repository: "Repository", commit: "Commit", path: Union[Path, str]
) -> Tuple["Repository", "Commit", Path]:
    """Resolve filename in submodules."""
    original_path = repository.path / path
    in_vendor = str(path).startswith(".renku/vendors")

    if original_path.is_symlink() or in_vendor:
        resolved_path = original_path.resolve()

        for submodule in repository.submodules:  # type: ignore
            if not (submodule.path / ".git").exists():
                continue

            try:
                path_within_submodule = resolved_path.relative_to(submodule.path)
                commit = submodule.get_previous_commit(path=path_within_submodule, revision=commit.hexsha)
            except (ValueError, errors.GitCommitNotFoundError):
                pass
            else:
                return submodule, commit, path_within_submodule

    return repository, commit, Path(path)


def get_dirty_paths(repository: "Repository") -> Set[str]:
    """Get paths of dirty files in the repository."""
    modified_files = [item.b_path for item in repository.unstaged_changes if item.b_path]
    staged_files = [d.a_path for d in repository.staged_changes] if repository.head.is_valid() else []

    return {os.path.join(repository.path, p) for p in repository.untracked_files + modified_files + staged_files}


@contextlib.contextmanager
def with_commit(
    *,
    repository: "Repository",
    transaction_id: str,
    commit_only=None,
    commit_empty=True,
    raise_if_empty=False,
    commit_message=None,
    abbreviate_message=True,
    skip_dirty_checks=False,
):
    """Automatic commit."""
    diff_before = prepare_commit(repository=repository, commit_only=commit_only, skip_dirty_checks=skip_dirty_checks)

    yield

    finalize_commit(
        diff_before=diff_before,
        repository=repository,
        transaction_id=transaction_id,
        commit_only=commit_only,
        commit_empty=commit_empty,
        raise_if_empty=raise_if_empty,
        commit_message=commit_message,
        abbreviate_message=abbreviate_message,
    )


def prepare_commit(*, repository: "Repository", commit_only=None, skip_dirty_checks=False, skip_staging: bool = False):
    """Gather information about repo needed for committing later on."""

    def ensure_not_untracked(path):
        """Ensure that path is not part of git untracked files."""
        for file_path in repository.untracked_files:
            is_parent = (repository.path / file_path).parent == (repository.path / path)
            is_equal = str(path) == file_path

            if is_parent or is_equal:
                raise errors.DirtyRenkuDirectory(repository)

    def ensure_not_staged(path):
        """Ensure that path is not part of git staged files."""
        path = str(path)
        for file_path in repository.staged_changes:
            is_parent = str(file_path.a_path).startswith(path)
            is_equal = path == file_path.a_path

            if is_parent or is_equal:
                raise errors.DirtyRenkuDirectory(repository)

    if skip_staging:
        if not isinstance(commit_only, list) or len(commit_only) == 0:
            raise errors.OperationError("Cannot use ``skip_staging`` without specifying files to commit.")

    diff_before = set()

    if commit_only == COMMIT_DIFF_STRATEGY:
        if len(repository.staged_changes) > 0 or len(repository.unstaged_changes) > 0:
            repository.reset()

        # Exclude files created by pipes.
        diff_before = {
            file for file in repository.untracked_files if STARTED_AT - int(Path(file).stat().st_ctime * 1e3) >= 1e3
        }

    if isinstance(commit_only, list) and not skip_dirty_checks:
        for path in commit_only:
            ensure_not_untracked(path)
            ensure_not_staged(path)

    return diff_before


def finalize_commit(
    *,
    diff_before,
    repository: "Repository",
    transaction_id: str,
    commit_only=None,
    commit_empty=True,
    raise_if_empty=False,
    commit_message=None,
    abbreviate_message=True,
    skip_staging: bool = False,
):
    """Commit modified/added paths."""
    from renku.core.util.urls import remove_credentials
    from renku.infrastructure.repository import Actor
    from renku.version import __version__, version_url

    committer = Actor(name=f"renku {__version__}", email=version_url)

    change_types = {item.a_path: item.change_type for item in repository.unstaged_changes}

    if commit_only == COMMIT_DIFF_STRATEGY:
        # Get diff generated in command.
        staged_after = set(change_types.keys())

        modified_after_change_types = {item.a_path: item.change_type for item in repository.staged_changes}

        modified_after = set(modified_after_change_types.keys())

        change_types.update(modified_after_change_types)

        diff_after = set(repository.untracked_files).union(staged_after).union(modified_after)

        # Remove files not touched in command.
        commit_only = list(diff_after - diff_before)

    if isinstance(commit_only, list):
        for path_ in commit_only:
            p = repository.path / path_
            if p.exists() or change_types.get(str(path_)) == DiffChangeType.DELETED:
                repository.add(path_)

    if not commit_only:
        repository.add(all=True)

    try:
        diffs = [d.a_path for d in repository.staged_changes]
    except errors.GitError:
        diffs = []

    if not commit_empty and not diffs:
        if raise_if_empty:
            raise errors.NothingToCommit()
        return

    if commit_message and not isinstance(commit_message, str):
        raise errors.CommitMessageEmpty()

    elif not commit_message:
        argv = [os.path.basename(sys.argv[0])] + [remove_credentials(arg) for arg in sys.argv[1:]]

        commit_message = " ".join(argv)

    if abbreviate_message:
        commit_message = shorten_message(commit_message)

    # NOTE: Only commit specified paths when skipping staging area
    paths = commit_only if skip_staging else []
    # Ignore pre-commit hooks since we have already done everything.
    repository.commit(commit_message + transaction_id, committer=committer, no_verify=True, paths=paths)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage interactive sessions.

Description
~~~~~~~~~~~

Manage sessions through the command line interface by starting and stopping them.
It's possible to list the current sessions and to connect to them.

Currently, two providers are supported: ``docker`` and ``renkulab``. More on this
later.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: special

.. click:: renku.ui.cli.session:session
   :prog: renku session
   :nested: full

Docker provider
~~~~~~~~~~~~~~~

The ``docker`` provider will take the current state of the repository, build a ``docker``
image (if one does not already exist) and then launch a session with this image. In
addition to this, the ``docker`` provider will mount the local repository inside
the ``docker`` container so that changes made in the session are immediately reflected
on the host where the session was originally started from.

Please note that in order to use this provider `Docker <https://docs.docker.com/>`_
is expected to be installed and available on your computer. In addition, using
this command from within a Renku interactive session started from the Renku website
is not possible. This command is envisioned as a means for users to quickly test
and check their sessions locally without going to a Renku deployment and launching
a session there, or in the case where they simply have no access to a Renku deployment.

.. code-block:: console

    $ renku session start -p docker

The command first looks for a local image to use. If a local image isn't found, it
searches the remote Renku deployment (if any) and pulls the image if it exists.
Finally, it prompts the user to build the image locally if no image is found. You
can force the image to always be built by using the ``--force-build`` flag.

This command accepts a subset of arguments of the ``docker run`` command. See
its help for the list of supported arguments: ``renku session start --help``.
Accepted values are the same as the ``docker run`` command unless stated
otherwise.

Renkulab provider
~~~~~~~~~~~~~~~~~

The ``renkulab`` provider will launch a regular interactive session
in the Renku deployment that hosts the current project. If the project has not
been uploaded/created in a Renku deployment then this provider will not be able
to launch a session. This provider is identical to going through the Renku website
and launching a session "manually" by selecting the project, commit, branch, etc.

Please note that there are a few limitations with the ``renkulab`` provider:

* If the user is not logged in (using the ``renku login`` command) then sessions
  can only be launched if the specific Renku deployment supports anonymous sessions.

* When launching anonymous sessions local changes cannot be reflected in them and
  changes made inside the session cannot be saved nor downloaded locally. This feature
  should be used only for adhoc exploration or work that can be discarded when
  the session is closed. The CLI will print a warning every time an anonymous session
  is launched.

* Changes made inside the interactive session are not immediately reflected locally,
  users should ``git pull`` any changes made inside an interactive session to get the
  same changes locally.

* Local changes can only be reflected in the interactive session if they are committed
  and pushed to the git repository. When launching a session and uncommitted or unpushed
  changes are present, the user will be prompted to confirm whether Renku should
  commit and push the changes before a session is launched. The session will launch
  only if the changes are committed and pushed.

.. code-block:: console

    $ renku session start -p renkulab

SSH connections to remote sessions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

You can connect via SSH to remote (Renkulab) sessions, if your project supports it.

To see if your project supports SSH, you can run ``renku project show`` and check the
``SSH Supported`` flag. If your project doesn't support SSH, update the project template
or contact the template maintainer to enable SSH support on the template.

You can start a session with SSH support using:

.. code-block:: console

    $ renku session start -p renkulab --ssh
    Your system is not set up for SSH connections to Renkulab. Would you like to set it up? [y/N]: y
    [...]
    Session <session-id> successfully started, use 'renku session open --ssh <session-id>' or 'ssh <session-id>' to
    connect to it

This will create SSH keys for you and setup SSH configuration for connecting to the renku deployment.
You can then use the SSH connection name (``ssh renkulab.io-myproject-sessionid`` in the example)
to connect to the session or in tools such as VSCode.

.. note::

   If you need to recreate the generated SSH keys or you want to use existing keys instead,
   you can use the ``renku session ssh-setup`` command to perform this step manually. See
   the help of the command for more details.

Alternatively, you can use ``renku session open --ssh <session_id>`` to directly open an SSH
connection to the session.

You can see the SSH connection name using ``renku session ls``.

SSH config for specific sessions is removed when the session is stopped.

Managing active sessions
~~~~~~~~~~~~~~~~~~~~~~~~

The ``session`` command can be used to also list, stop and open active sessions.
In order to see active sessions (from any provider) run the following command:

.. code-block:: console

    $ renku session ls -p renkulab
    ID                   STATUS    URL
    -------------------  --------  -------------------------------------------------
    renku-test-e4fe76cc  running   https://dev.renku.ch/sessions/renku-test-e4fe76cc

An active session can be opened by using its ``ID`` from the list above. For example, the ``open``
command below will open the single active session in the browser.

.. code-block:: console

    $ renku session open renku-test-e4fe76cc

An active session can be stopped by using the ``stop`` command and the ``ID`` from the list of
active sessions.

.. code-block:: console

    $ renku session stop renku-test-e4fe76cc

The command ``renku session stop --all`` will stop all active sessions regardless of the provider.

.. cheatsheet::
   :group: Managing Interactive Sessions
   :command: $ renku session start --provider renkulab
   :description: Start an interactive session on the remote Renku deployment.
   :target: rp

.. cheatsheet::
   :group: Managing Interactive Sessions
   :command: $ renku session ls
   :description: List all active sessions.
   :target: rp

.. cheatsheet::
   :group: Managing Interactive Sessions
   :command: $ renku session open <name>
   :description: Open a browser tab and connect to a running session.
   :target: rp

.. cheatsheet::
   :group: Managing Interactive Sessions
   :command: $ renku session stop <name>
   :description: Stop the specified session.
   :target: rp

"""

import click
from lazy_object_proxy import Proxy

from renku.command.format.session import SESSION_COLUMNS, SESSION_FORMATS
from renku.command.util import WARNING
from renku.core import errors
from renku.ui.cli.utils.callback import ClickCallback
from renku.ui.cli.utils.click import shell_complete_session_providers, shell_complete_sessions
from renku.ui.cli.utils.plugins import get_supported_session_providers_names


@click.group()
def session():
    """Session commands."""
    pass


@session.command("ls")
@click.option(
    "provider",
    "-p",
    "--provider",
    type=click.Choice(Proxy(get_supported_session_providers_names)),
    shell_complete=shell_complete_session_providers,
    default=None,
    help="Backend to use for listing interactive sessions.",
)
@click.option(
    "config",
    "-c",
    "--config",
    hidden=True,
    type=click.Path(exists=True, dir_okay=False),
    metavar="<config file>",
    help="YAML file containing configuration for the provider.",
)
@click.option(
    "--columns",
    type=click.STRING,
    default=None,
    metavar="<columns>",
    help="Comma-separated list of column to display: {}.".format(", ".join(SESSION_COLUMNS.keys())),
    show_default=True,
)
@click.option(
    "--format", type=click.Choice(list(SESSION_FORMATS.keys())), default="log", help="Choose an output format."
)
def list_sessions(provider, config, columns, format):
    """List interactive sessions."""
    from renku.command.session import session_list_command

    result = session_list_command().build().execute(provider=provider)

    click.echo(SESSION_FORMATS[format](result.output.sessions, columns=columns))

    if result.output.warning_messages:
        click.echo()
        if result.output.all_local and result.output.sessions:
            click.echo(WARNING + "Only showing sessions from local provider")
        for message in result.output.warning_messages:
            click.echo(WARNING + message)


def session_start_provider_options(*param_decls, **attrs):
    """Sets session provider options groups on the session start command."""
    from renku.core.plugin.session import get_supported_session_providers
    from renku.ui.cli.utils.click import create_options

    providers = [p for p in get_supported_session_providers() if p.get_start_parameters()]
    return create_options(providers=providers, parameter_function="get_start_parameters")


@session.command("start")
@click.option(
    "provider",
    "-p",
    "--provider",
    type=click.Choice(Proxy(get_supported_session_providers_names)),
    shell_complete=shell_complete_session_providers,
    default="docker",
    show_default=True,
    help="Backend to use for creating an interactive session.",
)
@click.option(
    "config",
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    metavar="<config file>",
    help="YAML file containing configuration for the provider.",
)
@click.option("--image", type=click.STRING, metavar="<image_name>", help="Docker image to use for the session.")
@click.option("--cpu", type=click.FLOAT, metavar="<cpu quota>", help="CPUs quota for the session.")
@click.option("--disk", type=click.STRING, metavar="<disk size>", help="Amount of disk space required for the session.")
@click.option(
    "--gpu",
    type=click.STRING,
    metavar="<GPU quota>",
    help="Number of GPU devices to add to the container ('all' to pass all GPUs).",
)
@click.option("--memory", type=click.STRING, metavar="<memory size>", help="Amount of memory required for the session.")
@session_start_provider_options()
def start(provider, config, image, cpu, disk, gpu, memory, **kwargs):
    """Start an interactive session."""
    from renku.command.session import session_start_command

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    communicator = ClickCallback()
    session_start_command().with_communicator(communicator).build().execute(
        provider=provider,
        config_path=config,
        image_name=image,
        cpu_request=cpu,
        mem_request=memory,
        disk_request=disk,
        gpu_request=gpu,
        **kwargs,
    )


@session.command("stop")
@click.argument("session_name", metavar="<name>", required=False, default=None, shell_complete=shell_complete_sessions)
@click.option(
    "provider",
    "-p",
    "--provider",
    type=click.Choice(Proxy(get_supported_session_providers_names)),
    default=None,
    shell_complete=shell_complete_session_providers,
    help="Session provider to use.",
)
@click.option("stop_all", "-a", "--all", is_flag=True, help="Stops all the running containers.")
def stop(session_name, stop_all, provider):
    """Stop an interactive session."""
    from renku.command.session import session_stop_command

    if stop_all and session_name:
        raise errors.ParameterError("Cannot specify a session ID with the '-a/--all' flag.")

    communicator = ClickCallback()
    session_stop_command().with_communicator(communicator).build().execute(
        session_name=session_name, stop_all=stop_all, provider=provider
    )
    if stop_all:
        click.echo("All running interactive sessions for this project have been stopped.")
    elif session_name:
        click.echo(f"Interactive session '{session_name}' has been successfully stopped.")
    else:
        click.echo("Interactive session has been successfully stopped.")


def session_open_provider_options(*param_decls, **attrs):
    """Sets session provider option groups on the session open command."""
    from renku.core.plugin.session import get_supported_session_providers
    from renku.ui.cli.utils.click import create_options

    providers = [p for p in get_supported_session_providers() if p.get_open_parameters()]  # type: ignore
    return create_options(providers=providers, parameter_function="get_open_parameters")


@session.command("open")
@click.argument("session_name", metavar="<name>", required=False, default=None, shell_complete=shell_complete_sessions)
@click.option(
    "provider",
    "-p",
    "--provider",
    type=click.Choice(Proxy(get_supported_session_providers_names)),
    shell_complete=shell_complete_session_providers,
    default=None,
    help="Session provider to use.",
)
@session_open_provider_options()
def open(session_name, provider, **kwargs):
    """Open an interactive session."""
    from renku.command.session import session_open_command

    session_open_command().build().execute(session_name=session_name, provider=provider, **kwargs)


@session.command("ssh-setup")
@click.option(
    "existing_key",
    "-k",
    "--existing-key",
    type=click.Path(exists=True, dir_okay=False),
    metavar="<private key file>",
    help="Existing private key to use.",
)
@click.option("--force", is_flag=True, help="Overwrite existing keys/config.")
def ssh_setup(existing_key, force):
    """Setup keys for SSH connections into sessions."""
    from renku.command.session import ssh_setup_command

    communicator = ClickCallback()
    ssh_setup_command().with_communicator(communicator).build().execute(existing_key=existing_key, force=force)
